from email import header


class response:
    """ A Response Object
    
    :rtype: Response
    """
    def __init__(self):
        """ Initialize the Response Object.

        :rtype: Response
        """
        self.status = "200"
        self.headers = {}
        self.body = "Hello From Retica"
        self.protocol = "HTTP/1.1"
        self.content_type = "text/html"
        self.content_length = len(self.body)
        self.connection = "close"

    def parse(self, string) -> None:
        """ Parse a response from a string and set the attributes of the response object.

            :param string: The Encoded Response String (Bytes).
            :type string: bytes

            :rtype: None
        """
        lines = string.split("\r\n")
        self.protocol, self.status = lines[0].split(" ")
        for line in lines[1:]:
            if line == "":
                break
            key, value = line.split(": ")
            self.headers[key] = value
        self.body = "\r\n".join(lines[len(lines)-1:])

    def compile(self) -> bytes:
        """ Compile the response into a string.
        
            :rtype: bytes
        """
        self.content_length = len(self.body)
        self.headers["Content-Type"] = str(self.headers["Content-Type"] if "Content-Type" in self.headers.keys() else self.content_type)
        self.headers["Content-Length"] = str(self.headers["Content-Length"] if "Content-Length" in self.headers.keys() else self.content_length)
        self.headers["Connection"] = str(self.headers["Connection"] if "Connection" in self.headers.keys() else self.connection)
        if type(self.body) == str:
            self.body = self.body.encode()
        return (self.protocol + " " + self.status + "\r\n" + "\r\n".join([key + ": " + value for key, value in self.headers.items()]) + "\r\n\r\n").encode() + self.body


def create(args):
    frontend_libraries = {
                            'React': [
                                '<script crossorigin src="https://unpkg.com/react@18/umd/react.development.js"></script>',
                                '<script crossorigin src="https://unpkg.com/react-dom@18/umd/react-dom.development.js"></script>',
                                '<script src="https://unpkg.com/babel-standalone@6/babel.min.js"></script>'
                            ],
                            'Vue': [
                                '<script src="https://unpkg.com/vue@3"></script>'
                            ],
                        }
    project_name = str(input("Project Name: ") or "Retica Project")
    project_description = str(input("Project Description: ") or "Retica Is A Performant, Easy To Use Web Server Based On Python With HTTPS & Plugins Support.")
    project_author = str(input("Project Author: ") or "CyroCoders")
    project_keywords = str(input("Project Keywords: ") or "Retica,Python,Web Server,Web Framework,CyroCoders")

    if input("Import Frontend Libraries? (y/n) ") == "y":
        for index, value in enumerate(frontend_libraries.keys()):
            print(f"{index}. {value}")
        frontend_libraries_selection = input("Select Frontend Library(Separate With ','. No Space): ")

    print("Creating project...")

    import tarfile
    import os
    
    file = tarfile.open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/default-homepage.tar.gz"))
    file.extractall(os.path.join(os.getcwd(), project_name))


    with open(os.path.join(os.getcwd(), project_name, "app.py"), "r") as in_data:
        in_data = in_data.readlines()
        
    with open(os.path.join(os.getcwd(), project_name, "app.py"), "w") as out_data:
        for line in in_data:
            if line.startswith("libraries = "):
                if frontend_libraries_selection:
                    out_data.write("libraries = {\n")
                    for lib in frontend_libraries_selection.split(","):
                        print(list(frontend_libraries.keys())[0])
                        out_data.write(f"   '{list(frontend_libraries.keys())[int(lib)]}': [\n")
                        for script in list(frontend_libraries.values())[int(lib)]:
                            out_data.write(f"       '{script}',\n")
                        out_data.write("   ],\n")
                    out_data.write("}\n")
                else:
                    out_data.write("libraries = {}\n")
            elif line.startswith("metadata = "):
                out_data.write("metadata = {\n")
                out_data.write(f"   'name': '{project_name}',\n")
                out_data.write(f"   'description': '{project_description}',\n")
                out_data.write(f"   'author': '{project_author}',\n")
                out_data.write(f"   'keywords': '{project_keywords}',\n")
                out_data.write("}\n")
            else:
                out_data.write(line)


import parse
from . import Sockets, Request, Response
import os, mimetypes
from typing import Union

class Worker:
    """ The Retica Worker

    :param server: The server to use.
    :type server: Server

    :rtype: Worker
    """
    def __init__(self, server):
        """ Initialize the Worker.

        :param server: The server to use.
        :type server: Server

        :rtype: Worker
        """
        self.server = server
    
    def handle(self,socket: Union[Sockets.HTTP_Socket,Sockets.HTTP_Socket],request: Request.request):
        """ Handle a request.

        :param socket: The socket to use.
        :type socket: Sockets.HTTP_Socket
        :param request: The request to handle.
        :type request: Request.request

        :rtype: None
        """
        try:
            response = Response.response()
            path = request.path
            path = path.decode()
            for plugin in self.server.plugins:
                plugin.intercept_request(request)
            for endpoint, (handler,condition) in self.server.endpoints.items():
                proceed = True
                if condition is not None:
                    proceed = condition(request)
                if proceed==True:
                    endpoint = "." + endpoint + "."
                    path = "." + path + "."
                    if endpoint[-1] == "/":
                        endpoint = endpoint[:-1]
                    if endpoint[1] == "/":
                        endpoint = endpoint[1:]
                    if path[-1] == "/":
                        path = path[:-1]
                    if path[1] == "/":
                        path = path[1:]
                    endpoint = endpoint[1:-1]
                    path = path[1:-1]
                    if parse.parse(endpoint, path):
                        arguments = parse.parse(endpoint, path).named
                        handler(request, response, **arguments)
                        for plugin in self.server.plugins:
                            plugin.intercept_response(response)
                        socket.sendall(response.compile())
                        return
                elif isinstance(proceed, Response.response):
                    response.body = proceed
                    socket.sendall(proceed.compile())
                    return
                elif isinstance(proceed, str):
                    response.body = proceed
                    socket.sendall(response.compile())
                    return
                elif isinstance(proceed, bytes):
                    response.body = proceed
                    socket.sendall(proceed.compile())
                    return

            static_path = os.path.join(os.path.dirname(os.path.abspath(self.server.context)), "static",*path.split("/"))
            if os.path.isfile(static_path):
                mimetype = mimetypes.guess_type(static_path)
                response.body = open(static_path, "rb").read()
                response.headers["Content-Type"] = mimetype[0] or "application/octet-stream"
                if mimetype[1]:
                    response.headers["Content-Encoding"] = mimetype[1]
                socket.sendall(response.compile())
                return
            else:
                response.status = "404"
                response.body = "404 Not Found"
                socket.sendall(response.compile())
                return
        except KeyboardInterrupt:
            pass
        except Exception as e:
            self.server.logger.exception(e)


import argparse

from .create import create
from .run import run

def main():
    parser = argparse.ArgumentParser(description="The Retica CLI. Retica Is A Performant, Easy To Use Web Server Based On Python With HTTPS & Plugins Support.")
    subparsers = parser.add_subparsers(dest='command')
    
    run_parser = subparsers.add_parser('run', help="Run the current project.")
    run_parser.add_argument("server_object", type=str, const="app:retica", default="app:retica", nargs='?', help="Server object (app:retica)")
    run_parser.add_argument("host", type=str, const="localhost:80", default="localhost:80", nargs='?', help="Hostname and port (localhost:80)")

    subparsers.add_parser('create', help="Create a new project.")

    args = parser.parse_args()
    match args.command:
        case 'create':
            create(args)
            return
        case 'run':
            run(args)
            return
        case None:
            parser.print_help()

if __name__ == "__main__":
    main()


from jinja2 import Environment, FileSystemLoader
import os

class TemplateRender:
    """ A Template Renderer
    
    :param server: The server to use.
    :type server: Server
    :param template_dir: The directory to use for templates.
    :type template_dir: str
    
    :rtype: TemplateRender
    """
    def __init__(self, server, template_dir="templates"):
        """ Initialize the Template Renderer.
        
        :param server: The server to use.
        :type server: Server
        :param template_dir: The directory to use for templates.
        :type template_dir: str
        
        :rtype: TemplateRender
        """
        self.template_dir = template_dir
        self.templateEnv = Environment(loader=FileSystemLoader(os.path.join(os.path.dirname(os.path.abspath(server.context)), template_dir)))

    def render(self, template_file, data=None):
        """ Render a template.

        :param template_file: The template file to use.
        :type template_file: str
        :param data: The data to use.
        :type data: dict

        :rtype: str
        """
        if data is None:
            data = {}
        return self.templateEnv.get_template(template_file).render(data)


from . import Worker, Request, Response, Plugin
import os, sys, select
import threading, multiprocessing
import logging

class Server:
    """ The Retica Server

    :param context: The server context to use.
    :type context: str
    :param logger: The logger to use.
    :type logger: logging.Logger
    :param config: The config to use.
    :type config: dict

    :rtype: Server
    """
    def __init__(self, context, logger: logging.Logger=None, config: dict=None):
        """ Initialize the Server.

        :param context: The server context to use.
        :type context: str
        :param logger: The logger to use.
        :type logger: logging.Logger
        :param config: The config to use.
        :type config: dict

        :rtype: Server
        """
        self.logger = logging.getLogger("SERVER") if logger is None else logger
        logging.basicConfig(level=logging.INFO)
        self.config = config
        self.context = context
        self.endpoints = {}
        self.workers = []
        self.plugins = []
        self.runner_plugins = []

    def use_plugin(self, plugin, config):
        """ Use a plugin.
        
        :param plugin: The plugin to use.
        :type plugin: Plugin.Plugin
        :param config: The config to use.
        :type config: dict
        
        :rtype: None"""
        if not issubclass(plugin, Plugin.Plugin):
            raise TypeError("Plugin must be of type Plugin")
        plugin_instance = plugin(self, config)
        self.plugins.append(plugin_instance)
        if hasattr(plugin, "run"):
            self.logger.info(f"Running Plugin {plugin.__name__}")
            self.runner_plugins.append(multiprocessing.Process(target=self.run_plugin, args=(plugin_instance,)).start())

    def run_plugin(self, plugin_instance):
        """ Run a plugin.
        
        :param plugin_instance: The plugin to run.
        :type plugin_instance: Plugin.Plugin
        
        :rtype: None
        """
        if not issubclass(plugin_instance.__class__, Plugin.Plugin):
            raise TypeError("Plugin must be of type Plugin")
        while True:
            plugin_instance.run()

    def create_endpoint(self, path, conditional=None):
        """ A Decorator to create an endpoint.
        
        :param path: The path to use.
        :type path: str
        :param conditional: The conditional to use.
        :type conditional: callable
        
        :rtype: callable"""
        def wrapper(handler):
            if(not(self.endpoints.__contains__(path))):
                self.endpoints[path] = (handler, conditional)
                return handler
            else:
                raise AssertionError(f"Endpoint {path}:{handler} Already Exists!")#self.error["urlcatcherexists"])
        return wrapper

    def run(self, sockets):
        """ Run the server.

        :param sockets: The list of sockets to use.
        :type sockets: list

        :rtype: None
        """
        self.logger.info("Starting Server")
        while True:
            try:
                for socket in sockets:
                    socket.bind(logger=self.logger)
                while True:
                    read_socks,_,_ = select.select(sockets, [], [], 0)
                    for sock in read_socks:
                        self.workers.append(threading.Thread(target=self.handle_request, args=(*sock.sock.accept(),)))
                        self.workers[-1].start()
                    read_socks = []
            except KeyboardInterrupt:
                self.logger.info("Stopping Server")
                for socket in sockets:
                    socket.close()
                for worker in self.workers:
                    worker.join()
                sys.exit(0)
            except Exception as e:
                self.logger.error(f"Error: {e}")
                self.logger.info("Restarting Server")

    def handle_request(self, socket, address):
        """ Handle a request.

        :param socket: The socket to use.
        :type socket: socket.socket
        :param address: The address to use.
        :type address: tuple
        
        :rtype: None
        """
        request = Request.request()
        request.parse(socket.recv(1024))
        self.logger.info(f"New Connection From {address} For {request.path.decode()}")
        worker = Worker.Worker(self)
        if not request:
            return
        worker.handle(socket, request)
        socket.close()


class Plugin(object):
    """ Base Plugin Class

    :param server: The server to use.
    :type server: Server
    :param config: The config to use.
    :type config: dict

    :rtype: Plugin
    """
    def __init__(self, server, config={}):
        """ Initialize the Plugin.
        
        :param server: The server to use.
        :type server: Server
        :param config: The config to use.
        :type config: dict
        
        :rtype: Plugin
        """
        self.server = server
        self.config = config

    def intercept_request(self, request):
        """ Intercept a request.
        
        :param request: The request to intercept.
        :type request: Request
        
        :rtype: Request
        """
        pass

    def intercept_response(self, response):
        """ Intercept a response.

        :param response: The response to intercept.
        :type response: Response

        :rtype: Response
        """
        pass


import socket
import ssl

def gethostname():
    """ Get the hostname of the current machine. """
    return socket.gethostname()

class HTTP_Socket(object):
    """ A HTTP Socket

    :param host: The host to bind to.
    :type host: str
    :param port: The port to bind to.
    :type port: int

    :rtype: HTTP_Socket
    """
    def __init__(self, host, port):
        """ Initialize the HTTP Socket.
        
        :param host: The host to bind to.
        :type host: str
        :param port: The port to bind to.
        :type port: int

        :rtype: HTTP_Socket
        """
        self.host = host
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def bind(self,logger=None):
        """ Bind the socket to the host and port.

        :param logger: The logger to use.
        :type logger: Logger

        :rtype: None
        """
        self.sock.bind((self.host, self.port))
        self.sock.listen(5)
        if logger:
            logger.info(f"Listening On {self.host}:{self.port}")
    
    def send(self, data):
        """ Send data to the socket.

        :param data: The data to send.
        :type data: bytes

        :rtype: None
        """
        self.sock.send(data)

    def recv(self, size):
        """ Receive data from the socket.

        :param size: The size of the data to receive.
        :type size: int

        :rtype: bytes
        """
        return self.sock.recv(size)

    def close(self):
        """ Close the socket.
    
        :rtype: None
        """
        self.sock.close()

    def fileno(self):
        """ Get the file descriptor of the socket.

        :rtype: int
        """
        return self.sock.fileno()
        return self.sock.fileno()

class HTTPS_Socket(HTTP_Socket):
    """ A HTTPS Socket
    
    :param host: The host to bind to.
    :type host: str
    :param port: The port to bind to.
    :type port: int
    :param cert_file: The certificate file to use.
    :type cert_file: str
    :param key_file: The key file to use.
    :type key_file: str

    :rtype: HTTPS_Socket
    """
    def __init__(self, host, port, cert_file, key_file):
        """ Initialize the HTTPS Socket.
        
        :param host: The host to bind to.
        :type host: str
        :param port: The port to bind to.
        :type port: int
        :param cert_file: The certificate file to use.
        :type cert_file: str
        :param key_file: The key file to use.
        :type key_file: str
        
        :rtype: HTTPS_Socket"""
        super().__init__(host, port)
        self.ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self.ssl_context.load_cert_chain(certfile=cert_file, keyfile=key_file)
        self.sock = self.ssl_context.wrap_socket(self.sock, server_side=True, do_handshake_on_connect=True)

    def bind(self):
        """ Bind the socket to the host and port.

        :param logger: The logger to use.
        :type logger: Logger

        :rtype: None
        """
        super().bind()
    
    def send(self, data):
        """ Send data to the socket.
        
        :param data: The data to send.
        :type data: bytes
        
        :rtype: None
        """
        super().send(data)

    def recv(self, size):
        """ Receive data from the socket.

        :param size: The size of the data to receive.
        :type size: int
        
        :rtype: bytes
        """
        super().recv(size)

    def close(self):
        """ Close the socket.

        :rtype: None
        """
        super().close()


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Retica",
    version="1.1.2",
    author="CyroCoders",
    author_email="pypi@cyrocoders.ml",
    description="Retica Is A Performant, Easy To Use Web Server Based On Python With HTTPS & Plugins Support.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CyroCoders/Retica",
    project_urls={
        "Bug Tracker": "https://github.com/CyroCoders/Retica/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    entry_points ={
        'console_scripts': [
            'retica = Retica.cli:main'
        ]
    },
    python_requires=">=3",
    install_requires=[
        'Jinja2==2.11.2',
        'Brotli==1.0.9',
        'parse==1.19.0',
        'pyCLI==2.0.3',
        'MarkupSafe==2.0.1'
    ]
)


class request:
    """ A Request Object
    
    :param method: The HTTP Method.
    :type method: str
    :param path: The path of the request.
    :type path: str
    :param protocol: The protocol of the request.
    :type protocol: str
    :param headers: The headers of the request.
    :type headers: dict
    :param data: The data of the request.
    :type data: dict
    :param body: The body of the request.
    :type body: bytes
    
    :rtype: Request
    """
    def __init__(self):
        """ Initialize the request object.

        :rtype: Request
        """
        self.method = None
        self.path = None
        self.protocol = None
        self.headers = {}
        self.data = {}
        self.body = None

    def parse(self, string: bytes) -> None:
        """ Parse a request from a string and set the attributes of the request object.
        
        :param string: The Encoded Request String (Bytes).
        :type string: bytes
        
        :rtype: None
        """
        lines = string.split(b"\r\n")
        self.method, self.path, self.protocol = lines[0].split(b" ")
        for line in lines[1:]:
            if line == b"":
                break
            key, value = line.split(b": ")
            self.headers[key] = value
        self.body = b"\r\n".join(lines[len(lines)-1:])

        self.data["GET"] = {
            key.decode(): value.decode()
            for key, value in [
                x.split(b"=") if type(x) != int else (f"{x}".encode(),f"{x}".encode()) for x in (
                    self.path.split(b"?")[1].split(b"&") if b"&" in self.path.split(b"?")[1] else self.path.split(b"?")[1]
                )
            ]
        } if b"?" in self.path else {}

        self.data["POST"] = {
            key.decode(): value.decode()
            for key, value in [
                x.split(b"=") for x in (
                    self.body.split(b"&") if b"&" in self.body else self.body
                )
            ]
        } if self.method.upper() == b"POST" else {}

        self.data["COOKIE"] = {
            key.decode(): value.decode()
            for key, value in [
                (x.split(b"=")[0],b"=".join(x.split(b"=")[1:])) for x in (
                    self.headers[b"Cookie"].split(b"; ") if b"; " in self.headers[b"Cookie"] else [self.headers[b"Cookie"]]
                )
            ]
        } if b"Cookie" in self.headers else {}

        self.path = self.path.split(b"?")[0] if b"?" in self.path else self.path
        
    def compile(self) -> bytes:
        """ Compile the request into a string.

            :rtpye: bytes
        """
        return self.method + b" " +  self.path + b"?" + b"&".join(["=".join([key, value]) for key, value in self.data["GET"].items()]) + b" " + self.protocol + b"\r\n" + b"\r\n".join([key + b": " + value for key, value in self.headers.items()]) + b"\r\n\r\n" + self.body


def run(args):
    import os.path
    import sys
    import importlib

    from Retica import Sockets

    sys.path.insert(0,os.getcwd())
    f,a = args.server_object.split(":")
    m = importlib.import_module(f)

    server = getattr(m, a)

    if __name__ == "Retica.cli.run":
        sock = Sockets.HTTP_Socket(args.host.split(":")[0], int(args.host.split(":")[1]))
        server.run([sock])